#include "tmesh.h"
#include <float.h>
#include "framebuffer.h"

#include <fstream>
#include <iostream>

using namespace std;

void TMesh::Initialize() {

  verts = 0;
  vertsN = 0;
  connectivity = 0;
  trianglesN = 0;
  projVerts = 0;
  normals = 0;
  colors = 0;

}

TMesh::TMesh() {

  Initialize();

}

void TMesh::SetAABB(AABB aabb) {

  vertsN = 8;
  trianglesN = 12;
  Allocate();

  int vi = 0;

  verts[vi] = V3(aabb.corners[0][0], aabb.corners[1][1],
    aabb.corners[1][2]);
  vi++;
  verts[vi] = V3(aabb.corners[0][0], aabb.corners[0][1],
    aabb.corners[1][2]);
  vi++;
  verts[vi] = V3(aabb.corners[1][0], aabb.corners[0][1],
    aabb.corners[1][2]);
  vi++;
  verts[vi] = V3(aabb.corners[1][0], aabb.corners[1][1],
    aabb.corners[1][2]);
  vi++;

  verts[vi] = V3(aabb.corners[0][0], aabb.corners[1][1],
    aabb.corners[0][2]);
  vi++;
  verts[vi] = V3(aabb.corners[0][0], aabb.corners[0][1],
    aabb.corners[0][2]);
  vi++;
  verts[vi] = V3(aabb.corners[1][0], aabb.corners[0][1],
    aabb.corners[0][2]);
  vi++;
  verts[vi] = V3(aabb.corners[1][0], aabb.corners[1][1],
    aabb.corners[0][2]);
  vi++;

  int ti = 0;
  connectivity[3*ti+0] = 0;
  connectivity[3*ti+1] = 1;
  connectivity[3*ti+2] = 2;
  ti++;
  connectivity[3*ti+0] = 2;
  connectivity[3*ti+1] = 3;
  connectivity[3*ti+2] = 0;
  ti++;
  connectivity[3*ti+0] = 3;
  connectivity[3*ti+1] = 2;
  connectivity[3*ti+2] = 6;
  ti++;
  connectivity[3*ti+0] = 6;
  connectivity[3*ti+1] = 7;
  connectivity[3*ti+2] = 3;
  ti++;
  connectivity[3*ti+0] = 7;
  connectivity[3*ti+1] = 6;
  connectivity[3*ti+2] = 5;
  ti++;
  connectivity[3*ti+0] = 5;
  connectivity[3*ti+1] = 4;
  connectivity[3*ti+2] = 7;
  ti++;
  connectivity[3*ti+0] = 0;
  connectivity[3*ti+1] = 4;
  connectivity[3*ti+2] = 5;
  ti++;
  connectivity[3*ti+0] = 0;
  connectivity[3*ti+1] = 1;
  connectivity[3*ti+2] = 5;
  ti++;
  connectivity[3*ti+0] = 0;
  connectivity[3*ti+1] = 3;
  connectivity[3*ti+2] = 7;
  ti++;
  connectivity[3*ti+0] = 7;
  connectivity[3*ti+1] = 4;
  connectivity[3*ti+2] = 0;
  ti++;
  connectivity[3*ti+0] = 1;
  connectivity[3*ti+1] = 5;
  connectivity[3*ti+2] = 6;
  ti++;
  connectivity[3*ti+0] = 6;
  connectivity[3*ti+1] = 2;
  connectivity[3*ti+2] = 1;
  ti++;

}

void TMesh::Allocate() {

  if (verts)
    delete []verts;
  verts = new V3[vertsN];

  if (projVerts)
    delete []projVerts;
  projVerts = new V3[vertsN];

  if (connectivity)
    delete []connectivity;
  connectivity = new unsigned int[3*trianglesN];

}

void TMesh::RenderPoints(V3 colorv, PPC *ppc, 
  FrameBuffer *fb) {

  int psize = 5;
  unsigned int color = colorv.GetColor();
  Project(ppc);
  for (int vi = 0; vi < vertsN; vi++) {
    if (projVerts[vi][0] == FLT_MAX)
      continue;
    int u = (int) projVerts[vi][0];
    int v = (int) projVerts[vi][1];
    fb->DrawPoint(u, v, psize, color);
  }

}


void TMesh::Project(PPC *ppc) {

  for (int vi = 0; vi < vertsN; vi++) {
    if (!ppc->Project(verts[vi], projVerts[vi])) {
      projVerts[vi] = V3(FLT_MAX, FLT_MAX, FLT_MAX);
    }
  }

}

void TMesh::RenderWireFrame(V3 colorv, PPC *ppc, 
  FrameBuffer *fb) {

  int psize = 5;
  unsigned int color = colorv.GetColor();
  Project(ppc);
  for (int ti = 0; ti < trianglesN; ti++) {
    for (int ei = 0; ei < 3; ei++) {
      int vi0 = connectivity[3*ti+ei];
      int vi1 = connectivity[3*ti+((ei+1)%3)];
      if (projVerts[vi0][0] == FLT_MAX)
        continue;
      if (projVerts[vi1][0] == FLT_MAX)
        continue;
      fb->DrawSegment(projVerts[vi0], projVerts[vi1], 
        color);
    }
  }

}

void TMesh::Load(char *fname) {

  ifstream ifs(fname, ios::binary);
  if (ifs.fail()) {
    cerr << "INFO: cannot open file: " << fname << endl;
    return;
  }

  ifs.read((char*)&vertsN, sizeof(int));
  char yn;
  ifs.read(&yn, 1); // always xyz
  if (yn != 'y') {
    cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
    return;
  }
  if (verts)
    delete verts;
  verts = new V3[vertsN];
  if (projVerts)
    delete projVerts;
  projVerts = new V3[vertsN];

  ifs.read(&yn, 1); // colors 3 floats
  if (colors)
    delete colors;
  colors = 0;
  if (yn == 'y') {
    colors = new V3[vertsN];
  }

  ifs.read(&yn, 1); // normals 3 floats
  if (normals)
    delete normals;
  normals = 0;
  if (yn == 'y') {
    normals = new V3[vertsN];
  }

  ifs.read(&yn, 1); // texture coordinates 2 floats
  float *tcs = 0; // don't have texture coordinates for now
  if (tcs)
    delete tcs;
  tcs = 0;
  if (yn == 'y') {
    tcs = new float[vertsN*2];
  }

  ifs.read((char*)verts, vertsN*3*sizeof(float)); // load verts

  if (colors) {
    ifs.read((char*)colors, vertsN*3*sizeof(float)); // load colors
  }

  if (normals)
    ifs.read((char*)normals, vertsN*3*sizeof(float)); // load normals

  if (tcs)
    ifs.read((char*)tcs, vertsN*2*sizeof(float)); // load texture coordinates

  ifs.read((char*)&trianglesN, sizeof(int));
  if (connectivity)
    delete connectivity;
  connectivity = new unsigned int[trianglesN*3];
  ifs.read((char*)connectivity, trianglesN*3*sizeof(unsigned int)); // read tiangles

  ifs.close();

  cerr << "INFO: loaded " << vertsN << " verts, " << trianglesN << " tris from " << endl << "      " << fname << endl;
  cerr << "      xyz " << ((colors) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;

}
